/* 
 * ppmtocpva
 * Convert from PPM to the PCL dialect used by the Citizen Printiva.
 * (c) J.C. Bradfield 2000.
 * Some parts based on ppmtopcl by Angus Duggan.
 * This program may be freely distributed, copied and modified.
 * Please rename significantly modified versions.
 * This program is not warranted for any purpose. It is experimental and 
 * may contain bugs of arbitrary severity. It may be possible to damage your
 * printer using this program.
 */

/* This is now of Windows driver quality with the dot screen.
   To celebrate this, we'll move to version 2. */

#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include <netpbm/pgm.h>
#include <netpbm/ppm.h>

static int highres = 0;
static int halftone = 0;
static int fsdith = 0;

/* This struct carries everything needed to do halftoning */
typedef struct {
  int cellsize; /* size of the dither matrix */
  int *dither; /* the dither matrix, as a simple array of length cellsize^2 */
  struct {
    int x,y,z; /* pythagorean triplet for screen angle: x positive */
    /* temporary storage for the incremental halftoning */
    int S1xi, S1xf, /* stage 1, x, integer and fractional */
		 S1yi, S1yf, /* ... y, ... */
		 S2xi, S2xf, S2yi, S2yf,
	       S3xi, S3xf, S3yi, S3yf,
      hrow,hcol ; /* the indices into the dither matrix */
  } C,M,Y,K;
} htinfo;

/* This macro returns the dither value for comp c in ht struct h */
#define ht_elt(h,c) ((h)->dither[(h)->cellsize*((h)->c.hrow)+((h)->c.hcol)])
/* this is the same but it transpose row and col */
#define ht_telt(h,c) ((h)->dither[(h)->cellsize*((h)->c.hcol)+((h)->c.hrow)])

/* The following is a rotation function, the incremental version
   of which is given by the following macros. The function is
   left here for reference */
/* same, but now take XYZ as Pyth rotn, and do the rotation using
   Paeth's three shear */
#if 0
void rotate(int row,int col,int *hrow,int *hcol,int x,int y,int z) {
  int i,j;
  /* first the x shear, with X' = (X-1)/Y = (x/z - 1)/(y/z) = (x-z)/y */
  i = 2*col*y + 2*row*(x-z); 
  if ( i >= 0 ) i = (i+y)/(2*y) ; else i = (i-y+1)/(2*y);
  j = row ;
  /* now the y shear, with Y' = Y = y/z */
  col = i;
  /* bloody rounding */
  row = 2*y*i + 2*j*z;
  if ( row >= 0 ) row = (row+z)/(2*z) ; else row = (row+1-z)/(2*z);
  /* now x shear again */
  i = 2*col*y + 2*row*(x-z); 
  if ( i >= 0 ) i = (i+y)/(2*y) ; else i = (i-y+1)/(2*y);
  j = row ;
  *hrow = j%cellsize; if ( *hrow < 0 ) (*hrow) += cellsize;
  *hcol = i%cellsize; if ( *hcol < 0 ) (*hcol) += cellsize;
}
#endif

/* This macro initializes the halftone structure for dithering
   component c
   at a given image row. The image col is assumed to be zero. */
#define ht_init(h,c,row) \
 (h)->c.S1xf = 2*row*((h)->c.x-(h)->c.z); \
 if ( (h)->c.y > 0 ) \
  (h)->c.S1xi = ((h)->c.S1xf-(h)->c.y+1)/(2*(h)->c.y); \
 else \
  (h)->c.S1xi = ((h)->c.S1xf+(h)->c.y+1)/(2*-(h)->c.y); \
 (h)->c.S1xf -= (h)->c.S1xi*2*(h)->c.y; \
 (h)->c.S1yi = row; (h)->c.S1yf = 0; \
 (h)->c.S2xi = (h)->c.S1xi; (h)->c.S2xf = 0; \
 (h)->c.S2yf = 2*(h)->c.y*(h)->c.S1xi + 2*(h)->c.z*(h)->c.S1yi; \
 if ( (h)->c.S2yf >= 0 ) (h)->c.S2yi = ((h)->c.S2yf+(h)->c.z)/(2*(h)->c.z) ; \
 else (h)->c.S2yi = ((h)->c.S2yf+1-(h)->c.z)/(2*(h)->c.z); \
 (h)->c.S2yf -= 2*(h)->c.z*(h)->c.S2yi ; \
 (h)->c.S3xf = 2*(h)->c.y*(h)->c.S2xi + 2*((h)->c.x-(h)->c.z)*(h)->c.S2yi ; \
 if ( (h)->c.S3xf >= 0 ) (h)->c.hcol = ((h)->c.S3xf+((h)->c.y>0?1:-1)*(h)->c.y)/(2*((h)->c.y>0?1:-1)*(h)->c.y); \
 else (h)->c.hcol = ((h)->c.S3xf-((h)->c.y>0?1:-1)*(h)->c.y+1)/(2*((h)->c.y>0?1:-1)*(h)->c.y); \
 (h)->c.S3xf -= 2*(h)->c.y*(h)->c.hcol; \
 (h)->c.hrow = (h)->c.S2yi ; (h)->c.S3yf = 0 ; \
 (h)->c.hrow %= (h)->cellsize ; if ( (h)->c.hrow < 0 ) (h)->c.hrow += (h)->cellsize ; \
 (h)->c.hcol %= (h)->cellsize ; if ( (h)->c.hcol < 0 ) (h)->c.hcol += (h)->cellsize ;

/* This macro increments the halftone calculation for component c
   by one column */
#define ht_inc(h,c) \
 /* stage 1: incrementing col means ... just */ \
 (h)->c.S1xi += ((h)->c.y>0?1:-1) ; \
 /* S1y is unchanged */ \
 /* Stage 2 : S1X has been incremented by sgn(y), so */ \
 (h)->c.S2xi += ((h)->c.y>0?1:-1) ; \
 (h)->c.S2yf += 2*(h)->c.y ; \
 if ( (h)->c.S2yf >= (h)->c.z ) { \
  /* increment S2y */ \
  (h)->c.S2yi++ ; (h)->c.S2yf -= 2*(h)->c.z; \
  /* now do stage three: S2x and y have both been incremented */ \
  (h)->c.hcol += ((h)->c.y>0?1:-1) ; /* because of the 2*y */ \
  (h)->c.S3xf += 2*((h)->c.x-(h)->c.z) ; \
  while ( (h)->c.S3xf < -((h)->c.y>0?1:-1)*(h)->c.y ) { (h)->c.hcol-- ; (h)->c.S3xf += 2*((h)->c.y>0?1:-1)*(h)->c.y ; } \
  (h)->c.hrow++; \
 } else if ((h)->c.S2yf >= -(h)->c.z ) { \
  /* no increment to S2y; do stage three with only S2x incd */ \
  (h)->c.hcol += ((h)->c.y>0?1:-1) ; \
 } else /* < -(h)->c.z */ { \
  /* decrement S2y */ \
  (h)->c.S2yi-- ; (h)->c.S2yf += 2*(h)->c.z; \
  /* now do stage three: S2x and y have both been decremented */ \
  (h)->c.hcol-- ; /* because of the 2*y */ \
  (h)->c.S3xf -= 2*((h)->c.x-(h)->c.z) ; \
  while ( (h)->c.S3xf >= ((h)->c.y>0?1:-1)*(h)->c.y ) { (h)->c.hcol++ ; (h)->c.S3xf -= 2*((h)->c.y>0?1:-1)*(h)->c.y ; } \
  (h)->c.hrow--; \
 } \
 if ( (h)->c.hrow >= (h)->cellsize ) (h)->c.hrow -= (h)->cellsize; \
 if ( (h)->c.hrow < 0 ) (h)->c.hrow += (h)->cellsize; \
 if ( (h)->c.hcol >= (h)->cellsize ) (h)->c.hcol -= (h)->cellsize; else \
 if ( (h)->c.hcol < 0 ) (h)->c.hcol += (h)->cellsize ;


/* This dithering algorithm is intended for use with the Printiva, which
   doesn't seem well-suited to simple Floyd-Steinberg.
   It is a clustered-dot ordered dither, with a few hacks.
   The standard monochrome dither takes a 4x4 cell, and replicates it into
   a 2x2 supercell: AB
                    CD
   where the cells A and D have black growing out from the centre, and
   the cells B and C have white growing out from the centre.
   In order to achieve the best (we hope) mix of black and colour, we
   do exactly that for CMY, but for black we swap the roles of AD and BC,
   so that black dots don't (usually...) coincide with colour dots.
   In addition, we offset the thresholds in D by one from A, out of a range
   of 64 instead of 32; so theoretically (at least) we double the resolution
   (and similarly for C and B).
   */


/* times 4, to get values from 256 */
static const int cluster8[8][8] = {
  { 24*4,12*4,20*4,26*4,39*4,51*4,43*4,37*4, },
  { 22*4, 0*4, 2*4, 4*4,41*4,63*4,61*4,59*4, },
  { 16*4, 8*4, 6*4,10*4,47*4,55*4,57*4,53*4, },
  { 30*4,18*4,14*4,28*4,33*4,45*4,49*4,35*4, },
  { 38*4,50*4,42*4,36*4,25*4,13*4,21*4,27*4, },
  { 40*4,62*4,60*4,58*4,23*4, 1*4, 3*4, 5*4, },
  { 46*4,54*4,56*4,52*4,17*4, 9*4, 7*4,11*4, },
  { 32*4,44*4,48*4,34*4,31*4,19*4,15*4,29*4 } } ;


/* here's an attempt at a highres table, 4 by 8 */
static const int hcluster[4][8] = {
  /* this is real */
  { 16*4, 0*4, 4*4,24*4,46*4,62*4,58*4,38*4, },
  { 28*4, 8*4,12*4,20*4,34*4,54*4,50*4,42*4, },
  { 44*4,60*4,56*4,36*4,18*4, 2*4, 6*4,26*4, },
  { 32*4,52*4,48*4,40*4,30*4,10*4,14*4,22*4  } };


/* try a line screen with 8x8 */
static const int dithmat8line[8*8] = {
    32*4,36*4,40*4,44*4,34*4,38*4,42*4,46*4,
     0*4, 4*4, 8*4,12*4, 2*4, 6*4,10*4,14*4,
    48*4,52*4,56*4,60*4,50*4,54*4,58*4,62*4,
    16*4,20*4,24*4,28*4,18*4,22*4,26*4,30*4,
    35*4,39*4,43*4,47*4,33*4,37*4,41*4,45*4,
     3*4, 7*4,11*4,15*4, 1*4, 5*4, 9*4,13*4,
    51*4,55*4,59*4,63*4,49*4,53*4,57*4,61*4,
    19*4,23*4,27*4,31*4,17*4,21*4,25*4,29*4 } ;


/* dot growing outward */
static const int dithmat6dot[6*6] = {
    20*4, 8*4,28*4,48*4,58*4,40*4,
     4*4, 0*4,12*4,60*4,62*4,56*4,
    32*4,16*4,24*4,36*4,52*4,44*4,
    50*4,60*4,42*4,22*4,10*4,30*4,
    62*4,64*4,58*4, 6*4, 2*4,14*4,
    38*4,54*4,46*4,34*4,18*4,26*4  };

/* dot growing outward, biased low */
static const int dithmat6dotlow[6*6] = {
    20*4, 8*4,28*4,50*4,63*4,40*4,
     4*4, 0*4,12*4,63*4,63*4,63*4,
    32*4,16*4,24*4,36*4,58*4,44*4,
    54*4,63*4,42*4,22*4,10*4,30*4,
    63*4,63*4,63*4, 6*4, 2*4,14*4,
    38*4,62*4,46*4,34*4,18*4,26*4  };

/* two lines, concentrating on the low end */
/* this more or less works with the same gamma
   setting as the dot screen, but it's inferior
   quality. I think there is no reason to continue 
   with this. */
static const int dithmat6linelow[6*6] = {
    54*5,30*5,42*5,46*5,24*5,33*5,
    22*5, 6*5,12*5,16*5, 0*5, 8*5,
    63*5,63*5,63*5,63*5,61*5,63*5,
    50*5,26*5,36*5,58*5,28*5,39*5,
    18*5, 2*5,10*5,20*5, 4*5,12*5,
    63*5,63*5,63*5,63*5,63*5,63*5 };

static const int dithmat10[10*10] = {
#if    0
    42*4,26*4,20*4,34*4,43*4,54*4,55*4,61*4,53*4,49*4,
    30*4,10*4, 4*4,14*4,40*4,56*4,58*4,63*4,59*4,57*4,
    18*4, 2*4, 0*4, 6*4,24*4,60*4,62*4,64*4,62*4,60*4,
    38*4,16*4, 8*4,12*4,32*4,57*4,59*4,63*4,58*4,56*4,
    44*4,36*4,22*4,28*4,45*4,48*4,52*4,61*4,55*4,47*4,
    54*4,55*4,61*4,53*4,49*4,42*4,27*4,21*4,35*4,43*4,
    56*4,58*4,63*4,59*4,57*4,31*4,11*4, 5*4,15*4,41*4,
    60*4,62*4,64*4,62*4,60*4,19*4, 3*4, 1*4, 7*4,25*4,
    57*4,59*4,63*4,58*4,56*4,39*4,17*4, 9*4,13*4,33*4,
    48*4,52*4,61*4,55*4,47*4,44*4,37*4,23*4,29*4,46*4
#else
    27*4,19*4,15*4,23*4,31*4,41*4,52*4,55*4,49*4,37*4,
    25*4,10*4, 4*4,12*4,21*4,43*4,58*4,62*4,60*4,48*4,
    17*4, 2*4, 0*4, 6*4,18*4,53*4,64*4,64*4,64*4,54*4,
    22*4,13*4, 8*4,14*4,26*4,47*4,61*4,63*4,59*4,45*4,
    33*4,24*4,16*4,20*4,29*4,35*4,50*4,56*4,51*4,39*4,
    42*4,52*4,55*4,49*4,38*4,28*4,19*4,15*4,23*4,32*4,
    44*4,58*4,62*4,60*4,48*4,25*4,11*4, 5*4,12*4,21*4,
    53*4,64*4,64*4,64*4,54*4,17*4, 3*4, 1*4, 7*4,18*4,
    47*4,61*4,63*4,59*4,46*4,22*4,13*4, 9*4,14*4,26*4,
    36*4,50*4,57*4,51*4,40*4,34*4,24*4,16*4,20*4,30*4,
#endif
} ;

static unsigned char initgamma[256] ;

/* new version for halftoning: start from no change except
   gamma and black adjustment */

#define NOHUE -1

/* Convert RGB to HSV, putting results in last three args;
   leaves hue unchanged if the result is no hue */
 void RGBtoHSV(r,g,b,hp,sp,vp)
     pixval r,g,b;
     double *hp,*sp,*vp;
{
  double rd, gd, bd, h, s, v, max, min, del, rc, gc, bc;

  /* convert RGB to HSV */
  rd = r / 255.0;            /* rd,gd,bd range 0-1 instead of 0-255 */
  gd = g / 255.0;
  bd = b / 255.0;

  /* compute maximum of rd,gd,bd */
  if (rd>=gd) { if (rd>=bd) max = rd;  else max = bd; }
  else { if (gd>=bd) max = gd;  else max = bd; }
  
  /* compute minimum of rd,gd,bd */
  if (rd<=gd) { if (rd<=bd) min = rd;  else min = bd; }
  else { if (gd<=bd) min = gd;  else min = bd; }

  del = max - min;
  v = max;
  if (max != 0.0) s = (del) / max;
  else s = 0.0;
  
  h = NOHUE;

  if (s != 0.0 && v != 0.0) {
    rc = (max - rd) / del;
    gc = (max - gd) / del;
    bc = (max - bd) / del;
    
    if      (rd==max) h = bc - gc;
    else if (gd==max) h = 2 + rc - bc;
    else if (bd==max) h = 4 + gc - rc;
    
    h = h / 6.0;
    if (h<0) h += 1;
  }
  
  /* map near-black to black to avoid weird effects */
  if (v <= .0625) s = 0.0;
  
  if ( h != NOHUE ) *hp = h ;
  *sp = s;
  *vp = v;
}

/* and vice versa */
 void HSVtoRGB(hf,sf,vf,rp,gp,bp)
     double hf,sf,vf;
     pixval *rp,*gp,*bp;
{
  int    j;
  double h, s, v, rd, gd, bd;
  double f, p, q, t;

  h = hf ; s = sf ; v = vf ;

  if (s==0.0) { rd = v;  gd = v;  bd = v; }
  else {
    if (h==1.0) h = 0.0;
    h = h * 6.0;
    j = (int) floor(h);
    f = h - j;
    p = v * (1-s);
    q = v * (1 - (s*f));
    t = v * (1 - (s*(1 - f)));
    
    switch (j) {
    case 0:  rd = v;  gd = t;  bd = p;  break;
    case 1:  rd = q;  gd = v;  bd = p;  break;
    case 2:  rd = p;  gd = v;  bd = t;  break;
    case 3:  rd = p;  gd = q;  bd = v;  break;
    case 4:  rd = t;  gd = p;  bd = v;  break;
    case 5:  rd = v;  gd = p;  bd = q;  break;
    }
  }

  *rp = (pixval) floor(rd * 255);
  *gp = (pixval) floor(gd * 255);
  *bp = (pixval) floor(bd * 255);
}



void jcb_cmyk_ht(cp,mp,yp,kp)
     int *cp,*mp,*yp,*kp;
{
  int c,m,y,k,hmax,knew;
  double h,s,v;
  c = *cp; m = *mp ; y = *yp ; k = *kp ;
  /* add the black back into the colours */ 
  c += k ; m += k ; y += k ;  
  /* first adjust the val by highly non-linear factor */
  RGBtoHSV(255-c,255-m,255-y,&h,&s,&v);
  v = 1.0-initgamma[(int)((1.0-v)*255.0)]/255.0;
  s = (3*initgamma[(int)(s*255.0)]/255.0 + 2*s)/5.0;
  HSVtoRGB(h,s,v,&c,&m,&y);
  c = 255 - c; m = 255 - m; y = 255 - y;
  m = 97*m/100;
  k = c ; if ( m < k ) k = m ; if ( y < k ) k = y ; 
  hmax = c; if ( m > hmax ) hmax = m ; if ( y > hmax ) hmax = y;
  /* set k = big for highly sat, k = low for low sat */
  /* rethinking: we want a lot of black left in colours
     that are dark *or* highly saturated */
  /*
    Let's say we want almost full black for colours darker
    than about 50%.
  */
  knew = k/2; k = 3*k/4 ;
  c -= knew ; m -= knew ; y -= knew ;
  if ( fsdith ) { c = 2*c/3 ; m = 2*m/3; y = 2*y/3 ; k = 2*k/3 ; }
  *cp = c ; *mp = m ; *yp = y ; *kp = k ;
#if 0
   assert(c >=0 && c <= 255);
   assert(m >=0 && m <= 255);
   assert(y >=0 && y <= 255);
   assert(k >=0 && k <= 255);
#endif
}

/* compare integers for sorting */
static int intcomp(const void *a,const void *b) {
  return *(int *)a - *(int *)b;
}

/* Function for building dither matrices.
   indith is an n by n dither matrix with values in 0-255.
   condith is an m by m matrix with values in 0..m*m-1
   The output is an mn by mn matrix, in which the [i,j]th
   element of the [p,q]th n by n block has a value which is
   condith[p,q]/m*m of the way between indith[i,j] and the 
   next higher value (ignoring duplicates) that occurs in indith.
   The highest value is interpolated to 256.
   To ensure that solid colour is always printed as such,
   values in the out matrix are clamped to 254. */
int *build_dith(int n, const int *indith, int m, const int *condith) {
  int *sorted;
  int *result;
  int i,j,p,q;
  /* build a sorted copy of indith */
  sorted = (int *)malloc(sizeof(int)*n*n);
  memcpy(sorted,indith,sizeof(int)*n*n);
  qsort(sorted,n*n,sizeof(int),intcomp);

  result = (int *)malloc(sizeof(int)*m*n*m*n);
  for ( i = 0; i < n; i++ ) {
    for ( j = 0; j < n; j++ ) {
      int val, nval, k;
      /* find the value in sorted order */
      val = indith[i*n+j];
      for ( k = 0; sorted[k] != val ; k++ );
      /* get the next value */
      while ( sorted[k] == val && k < n*n ) k++;
      if ( k == n*n ) nval = 256 ; else nval = sorted[k];
      /* and interpolate */
      for ( p = 0; p < m; p++ ) {
	for ( q = 0; q < m; q++ ) {
	  int res = (int)rint(((m*m-condith[p*m+q])*val
		       +condith[p*m+q]*nval)/(1.0*m*m));
	  if ( res > 254 ) res = 254;
	  result[(n*p+i)*m*n + (n*q + j)] = res;
	}
      }
    }
  }
  return result;
}



#define MAXCOLORS 1024

extern double atof();
extern int atoi();

/* dot weight */
static double cdot = 1.0, mdot = 1.0, ydot = 1.0, kdot = 1.0;
static double fthresh = 0.5 ;
static int icdot, imdot, iydot, ikdot, ithresh; /* integer versions */

char *usage=" [-lowres|-highres [-highresinput]]\n\
              [-banding]\n\
              [-black]\n\
              [-coldiag]\n\
              [-curlcorrection|-nocurl]\n\
              [-dummy <ignored-arg>]\n\
              [-evenrowsonly|-oddrowsonly]\n\
              [-fine]\n\
              [-firstplane|-midplane|-lastplane]\n\
              [-fsdither|-halftone|-squaredither]\n\
              [-fullblack|-nocolcor|-noblack]\n\
              [-greyfile <pgm file>]\n\
              [-htmx <X> -htmy <Y> -htmz <Z>]\n\
              [-initgamma <gamma factor>]\n\
              [-lfadj <adjustment>]\n\
              [-media <media type>]\n\
              [-mono]\n\
              [-nopack]\n\
              [-noreset]\n\
              [-pageC] [-pageK] [-pageM] [-pageY]\n\
              [-pagelength <length>]\n\
              [-pagewidth <width>]\n\
              [-papersize <paper size>]\n\
              [-phadj <adjustment>]\n\
              [-ppmdiag|-ppmout]\n\
              [-raster]\n\
              [-solidblack]\n\
              [-spotfollows]\n\
              [-spotmode]\n\
              [-version]\n\
              [-xshift <offset>]\n\
              [-yshift <offset>]\n";
              

/* The actual maximum is 4800, but since the printer discards excess data,
   there's no point in creating errors for small overruns */
#define PCL_MAXWIDTH 6000
#define PCL_MAXHEIGHT 10000 /* 6375 is the A4 value. */
#define PCL_MAXVAL 255

static int nopack = 0;
static int xshift = 0;
static int yshift = 0;
static char dummyarg[512];
static int pagelength = 0;
static int pagewidth = 0;
static int noreset = 0;
static int ffonly = 0; /* do formfeed at end of page, but not reset */
static int alreadyhigh = 0; /* the incoming data is already 1200x600 */
static int lowres = 0;
static int raster = 0; /* forces the use of raster mode */
static int spotfollows = 0;
/* Spotmode: if -page flag given, print in 4 colour spot mode, otherwise 
   in 4 colour raster mode; in either case C, M, Y, K in the input corresponds
   to cartridge 1,2,3,4 */
static int spotmode = 0 ;

static int black = 0;
static int noblack = 0;
static int nocurl = 0;
static int fullblack = 0 ; /* full black generation -- implies nocolcor */
static int solidblack = 0; /* map pure black to pure black */
static int nocolcor = 0;
static int ppmout = 0;
static int coldiag = 0;
static int ppmdiag = 0;
static int mono = 0;
static int banding = 0;
static double initgam = 0.0;
static int oddrowsonly = 0;
static int evenrowsonly = 0;
static int curlcorrection = 0;
static int fine = 0;
static int lfadj = 0 ;
static int phadj = 0;
static int testint = 0 ;
static int dithering = 0;
static char greyfile[512];
static char mediatype[64];
static int mediacode;
static int finecode;
static char papersize[64];
static int version;
static int papercode;
static int htmx,htmy,htmz;
static int pageC = 0, pageM = 0 , pageY = 0 , pageK = 0, pageplane = 0;
/* By default, in pageplane mode C is assumed to be first and K last.
   This can be overridden by the firstplane, midplane and lastplane options
   */
static int lastplane = 0, firstplane = 0, midplane = 0;

/* argument types */
#define DIM 0
#define REAL 1
#define BOOL 2
#define INT  3
#define STRING 4
#define HDIM 5 /* horizontal dimension */
static struct options {
   char *name;
   int type;
   char *value;
} options[] = {
   {"banding",     BOOL, (char*)&banding },
   {"black",        BOOL, (char *)&black },
   {"coldiag",        BOOL, (char *)&coldiag },
   {"curlcorrection", BOOL, (char *)&curlcorrection},
   {"dummy", STRING, (char *)dummyarg },
   {"evenrowsonly",          BOOL, (char *)&evenrowsonly},
   {"ffonly",BOOL, (char *)&ffonly},
   {"fine",            BOOL, (char *)&fine},
   {"firstplane",   BOOL, (char *)&firstplane },
   {"fsdither",     BOOL, (char *)&fsdith },
   {"fullblack",     BOOL, (char *)&fullblack },
   {"greyfile",     STRING, (char *)greyfile },
   {"halftone",     BOOL, (char *)&halftone},
   {"highres",      BOOL, (char *)&highres },
   {"highresinput", BOOL, (char*)&alreadyhigh },
   {"htmx", INT, (char*)&htmx},
   {"htmy", INT, (char*)&htmy},
   {"htmz", INT, (char*)&htmz},
   {"initgamma",    REAL, (char *)&initgam },
   {"lastplane",   BOOL, (char *)&lastplane },
   {"lfadj",           INT,  (char *)&lfadj},
   {"lowres",      BOOL, (char *)&lowres },
   {"media",      STRING, (char *)mediatype },
   {"midplane",   BOOL, (char *)&midplane },
   {"mono",          BOOL, (char *)&mono},
   {"noblack",      BOOL, (char *)&noblack },
   {"nocolcor",     BOOL, (char *)&nocolcor },
   {"nocurl",      BOOL, (char *)&nocurl },
   {"nopack",       BOOL, (char *)&nopack },
   {"noreset",      BOOL, (char *)&noreset },
   {"oddrowsonly",          BOOL, (char *)&oddrowsonly},
   {"pageC",        BOOL, (char *)&pageC },
   {"pageK",        BOOL, (char *)&pageK },
   {"pageM",        BOOL, (char *)&pageM },
   {"pageY",        BOOL, (char *)&pageY },
   {"pagelength",   DIM, (char *)&pagelength },
   {"pagewidth",    HDIM, (char *)&pagewidth },
   {"papersize",    STRING, (char *)papersize },
   {"phadj",           INT,  (char *)&phadj},
   {"ppmdiag",             BOOL, (char *)&ppmdiag},
   {"ppmout",             BOOL, (char *)&ppmout},
   {"raster",       BOOL, (char *)&raster },
   {"solidblack",      BOOL, (char *)&solidblack },
   {"spotfollows",      BOOL, (char *)&spotfollows },
   {"spotmode",      BOOL, (char *)&spotmode },
   {"squaredither",    BOOL, (char *)&dithering },
   {"testint",     INT,  (char *)&testint}, /* random int for testing purposes */
   {"version", BOOL, (char*)&version },
   {"xshift",       HDIM,  (char *)&xshift },
   {"yshift",       DIM,  (char *)&yshift },
};

#define putword(w) (putchar(((w)>>8) & 255), putchar((w) & 255))

int bitsperpixel(v)
     int v;
{
   int bpp = 0;
   while (v > 0) {  /* calculate # bits for value */
      ++bpp;
      v>>=1;
   }
   return (bpp);
}

static char *inrow = NULL;
static char *outrow = NULL;
static int numrows = 4 ;
static int rows ;
static /*signed*/ char *runcnt = NULL;

void putbits(bp,n) /* take the char array bp, size n, and compress and output */
     char *bp; int n;
{
   int out = 0;
   int num = 0;
   static int rownum = 0 ; /* flush plane on row 4 */
   static int rowstoskip = 0; /* blank rows pending */
   static int pack = -1; /* to force initialization of compression state */
   int i, bitno;
   static char rastercommand;

   if (ppmout) return;

   if ( rownum == 0 ) rowstoskip = 0;

   /* first pack the byte array into bits */
   for ( i = 0, bitno = 0 ; i < n; i++, bitno++,bp++ ) {
     if ( bitno == 0 ) out = 0;
     if ( *bp ) out |= (1 << (7 - bitno));
     if ( bitno == 7 || i == n-1) {
       inrow[num++] = out;
       bitno -= 8;
     }
   }
   
   /* now pack the row */
   for (; num > 0 && inrow[num-1] == 0; num--); /* remove trailing zeros */
   if (num && !nopack) {            /* TIFF 4.0 packbits encoding */
     int start = 0;
     int next;
     runcnt[start] = 0;
     for (i = 1; i < num; i++) {
       if (inrow[i] == inrow[i-1]) {
	 if (runcnt[start] <= 0 && runcnt[start] > -127)
	   runcnt[start]--;
	 else
	   runcnt[start = i] = 0;
       } else {
	 if (runcnt[start] >= 0 && runcnt[start] < 127)
	   runcnt[start]++;
	 else
	   runcnt[start = i] = 0;
       }
     }
     start = 0;
     for (i = 0; i < num; i = next) {
       int count = runcnt[i];
       int from = i;
       if (count >= 0) { /* merge two-byte runs */
	 for (;;) {
	   next = i+1+runcnt[i];
	   if(next >= num || runcnt[next] < 0 ||
	      count+runcnt[next]+1 > 127)
	     break;
	   count += runcnt[next]+1;
	   i = next;
	 }
       }
       next =  i + 1 + ((runcnt[i] < 0) ? -runcnt[i] : runcnt[i]);
#if 0
       /* I cannot for the life of me work out what this piece
	  of code is supposed to be doing. Cutting it out seems
	  to have no deleterious effects, and so far it also seems
	  to eliminate the mysterious lines that appeared occasionally.
	  Where did I nick this code from in the first place?
       */
       if (next < num && count > 0 &&
	   runcnt[next] < 0 && runcnt[next] > -127) {
	 count--;
	 next--;
	 runcnt[next] = runcnt[next+1]-1;
       }
#endif
       outrow[start++] = count;
       if (count >= 0) {
	 while (count-- >= 0)
	   outrow[start++] = inrow[from++];
       } else
	 outrow[start++] = inrow[from];
     }
     if (start < num) {
       num = start;
       if (pack < 1) {
	 printf("\033*b\002%cM",0); /* select TIFF */
       }
       pack = 1;
     } else {
       if (pack != 0) {
	 printf("\033*b%c%cM",0,0); /* select uncompressed */
       }
       pack = 0;
     }
   }
   /* In raster mode, the first three colours are transferred with
      V, the fourth with W. In pagemode, we always use W, except for the
      very last row transferred. */
   /* first, if we're in 4 color raster mode, we have to select the cartridge (sigh) */
   if ( ! pageplane && spotmode ) {
     /* normal raster data is sent in the order K C M Y; whereas our convention
        is to have the cartridges in the order C M Y K. Hence the little calculation */
     printf("\033\032%c%cc",(rownum==0)?3:(rownum-1),nocurl?0:1);
   }
   if ( ! pageplane ) {
     rastercommand = (++rownum == numrows)?(rownum = 0, 'W') : 'V' ;
   } else {
     rastercommand = (++rownum == rows)? 'V' : 'W' ;
   }
   /* in pageplane mode, we'll skip blank rows rather than print them, to avoid
      advancing the paper unnecessarily. We ought to do this in raster mode as well,
      but then it's more complicated.
      Don't do this if no reset, or nothing will happen */
   if ( pageplane && num == 0 && ! noreset) { 
     rowstoskip++ ;
   } else {
     if ( rowstoskip ) printf("\033*b%c%cY",rowstoskip%256,rowstoskip/256);
     rowstoskip = 0;
     printf("\033*b%c%c%c", num%256,num/256,rastercommand);
     for (i = 0; i < num; i++)
       putchar(pack ? outrow[i] : inrow[i]);
   }
   num = 0; /* new row */
}


/* These are the values we'll use for the rotations */


#if 1
#define HT_C_X 63
#define HT_C_Y -16
#define HT_C_Z 65
#define HT_M_X 99
#define HT_M_Y 20
#define HT_M_Z 101
#define HT_K_X 4
#define HT_K_Y 3
#define HT_K_Z 5
#define HT_Y_X 63
#define HT_Y_Y 16
#define HT_Y_Z 65
#else
#define HT_C_X 63
#define HT_C_Y -16
#define HT_C_Z 65
#define HT_M_X 63
#define HT_M_Y 16
#define HT_M_Z 65
#define HT_K_X 4
#define HT_K_Y 3
#define HT_K_Z 5
#define HT_Y_X 4
#define HT_Y_Y 3
#define HT_Y_Z 5
#endif

htinfo kht =
{
  0,
  0,
  { HT_C_X, HT_C_Y, HT_C_Z, 0,0,0,0,0,0,0,0,0,0,0,0,0,0 },
  { HT_M_X, HT_M_Y, HT_M_Z, 0,0,0,0,0,0,0,0,0,0,0,0,0,0},
  { HT_Y_X, HT_Y_Y, HT_Y_Z, 0,0,0,0,0,0,0,0,0,0,0,0,0,0},
  { HT_K_X, HT_K_Y, HT_K_Z, 0,0,0,0,0,0,0,0,0,0,0,0,0,0}
};


int main(argc, argv)
     int argc;
     char *argv[];
{
   FILE *ifd, *gfd;
   register pixel *pixels, *pixrow;
   gray *gpixels,*gpixrow;
   unsigned char *blackbits, *cyanbits, *magentabits, *yellowbits;
   unsigned char *kbp, *cbp, *mbp, *ybp;
   register int  row, col, bpp, i;
   int grows, gcols, gmaxval, gformat;
   int cols, format; /* for sordid reasons (see putbits) rows is now
			top level */
   pixval maxval;
   long* thiscerr;
   long* nextcerr;
   long* thismerr;
   long* nextmerr;
   long* thisyerr;
   long* nextyerr;
   long* thiskerr;
   long* nextkerr;
   long* temperr;
   register long sc, sm, sy, sk, err,nc,nm,ny,nk,ac,am,ay,ak;
#define FS_BASIC_SCALE 16
#define FS_SCALE (FS_BASIC_SCALE*maxval)
   int fs_direction;
   int rowcounter;
   int index;
   int r,g,b,c,m,y,k;
   int ic,im,iy;
   int skipcols;
   int printrow;
   int bandingoffset;
   int zero = 0;
   int hrhalf, hrtoggle ; /* variables used in highres halftoning */
   unsigned char *colconv; /* huge look up table */
   
   ppm_init( &zero, argv ); /* don't let pbmplus at the arguments */
   /*   ppm_init( &argc, argv );*/

   *greyfile = '\000';

   while (argc > 1 && argv[1][0] == '-') {
      char *c, *val=argv[1]+1;
      for (i = 0; i < (int)(sizeof(options)/sizeof(struct options)); i++) {
	 if (strlen(val) <= strlen(options[i].name) &&
	     !strncmp(val, options[i].name, strlen(val))) {
	    switch (options[i].type) {
	      int resf; /* the dimension parsing assumes
				300dpi, but our normal resolution
				is 600 dpi.
				This assumes that the highres or lowres
				options have already been given.
			     */
	    case HDIM:
	    case DIM:
	      resf = 2; /* default is 600 dpi */
	      if ( lowres ) resf = 1;
	      if ( highres && options[i].type == HDIM) resf *= 2;
	       if (++argv, --argc == 1)
		  pm_usage(usage);
	       for (c = argv[1]; isdigit(*c) || *c == '.' || *c == '-'; c++);
	       if (c[0] == 'p' && c[1] == 't') /* points */
		  *(int *)(options[i].value) = (int)(resf*atof(argv[1])*4.16667);
	       else if (c[0] == 'd' && c[1] == 'p') /* decipoints */
		  *(int *)(options[i].value) = (int)(resf*atof(argv[1])*0.41667);
	       else if (c[0] == 'i' && c[1] == 'n') /* inches */
		  *(int *)(options[i].value) = (int)(resf*atof(argv[1])*300);
	       else if (c[0] == 'c' && c[1] == 'm') /* centimetres */
		  *(int *)(options[i].value) = (int)(resf*atof(argv[1])*118.11024);
	       else if (c[0] == 'm' && c[1] == 'm') /* millimetres */
		  *(int *)(options[i].value) = (int)(resf*atof(argv[1])*11.81102);
	       else if (!c[0]) /* dots */
		 /* NB here we assume the user knows what it's doing */
		  *(int *)(options[i].value) = atoi(argv[1]);
	       else
		  pm_error("illegal unit of measure %s", c);
	       break;
	    case REAL:
	       if (++argv, --argc == 1)
		  pm_usage(usage);
	       *(double *)(options[i].value) = atof(argv[1]);
	       break;
	    case BOOL:
	       *(int *)(options[i].value) = 1;
	       break;
	     case INT:
	       if (++argv, --argc == 1)
		  pm_usage(usage);
	       *(int *)(options[i].value) = atoi(argv[1]);
	       break;
	    case STRING:
	      if (++argv, --argc == 1)
		pm_usage(usage);
	      strcpy(options[i].value,argv[1]);
	    }
	    break;
	 }
      }
      if (i >= (int)(sizeof(options)/sizeof(struct options)))
	 pm_usage(usage);
      argv++; argc--;
   }
   if (argc > 2)
      pm_usage(usage);
   else if (argc == 2)
      ifd = pm_openr(argv[1]);
   else
      ifd = stdin ;

   if ( version ) {
     fprintf(stderr,"ppmtocpva version 1.0, 2001-08-04\n");
     exit(0);
   }
   /* validate arguments */

   if ( ppmdiag || coldiag ) ppmout = 1 ;

   if ( fsdith ) dithering = 1;
   if (halftone) dithering = 1;

   if ( black && noblack )
     pm_error("only one of -black and -noblack may be used");

   /* if none of the page flags are set, we're in
      raster mode. If one is set, we assume pageplane (unless explicit
      raster option given).
      If more than one, we print in raster, but only print
      the given colours */
   
   if ( black ) { pageK = 1 ; pageC = pageM = pageY = 0 ; }

   if (pageC + pageM + pageY + pageK == 1 && ! black && !raster) { pageplane = 1 ; }
   else if (pageC + pageM + pageY + pageK == 0 )
     pageC = pageM = pageY = pageK = 1 ;

   if (pageplane && ! firstplane && ! lastplane && !midplane) {
     firstplane = pageC;
     lastplane = pageK;
   }

   /* mediatype. These are values given by the PCL spec */
   mediacode = 0; /* default paper */
   if ( mediatype && mediatype[0] ) {
     if ( strcmp(mediatype,"paper") == 0 )
       mediacode = 0 ; 
     if ( strcmp(mediatype,"cold") == 0 )
       mediacode = 1 ; /* cold transparency */
     else if ( strcmp(mediatype,"highgrade") == 0 )
       mediacode = 7 ; /* high grade paper */
     else if ( strcmp(mediatype,"thick") == 0 )
       mediacode = 5 ; /* thick paper or card */
     else if ( strcmp(mediatype,"ohp") == 0 )
       mediacode = 8 ; /* transparency */
     else if ( strcmp(mediatype,"label") )
       mediacode = 3 ; /* label sheet (unsup) */
     else if ( strcmp(mediatype,"iron-on") )
       mediacode = 2 ; /* iron-on sheet */
     else if ( strcmp(mediatype,"postcard") )
       mediacode = 6; 
     else if ( strcmp(mediatype,"backprint") )
       mediacode = 9;
     else pm_error("unknown media type");
   }
   if ( mediacode == 9 || mediacode == 7 || mediacode == 6 
	|| mediacode == 5 || mediacode == 3 || mediacode == 2 ) {
     if ( fine ) pm_error("-fine cannot be used with this media type");
     finecode = 0;
     if ( mediacode == 2 ) finecode = 2 ; /* iron-on needs PPC */
   } else {
     finecode = 1;
     if ( fine ) finecode = 2;
   }

   /* papersize */
   papercode = 4 ; /* default to a4 */
   if ( papersize && papersize[0] ) {
     if ( strcmp(papersize,"custom") ) papercode = 0;
     else if ( strcmp(papersize,"user") ) papercode = 0;
     else if ( strcmp(papersize,"executive") ) papercode = 1;
     else if ( strcmp(papersize,"letter") ) papercode = 2;
     else if ( strcmp(papersize,"legal") ) papercode = 3;
     else if ( strcmp(papersize,"a4") ) papercode = 4;
     else if ( strcmp(papersize,"A4") ) papercode = 4;
     else if ( strcmp(papersize,"b5") ) papercode = 5;
     else if ( strcmp(papersize,"B5") ) papercode = 5;
     else if ( strcmp(papersize,"postcard") ) papercode = 6;
     else pm_error("unknown paper size");
   }
   /* it appears that the papersize command does not actually set
      the logical page length and width!  How very stupid.
      So we do it ourselves, unless the user specified it */
   if ( pagelength == 0 ) {
     int l[] = { 0, 5663, 5962, 7755, 6375, 5434, 2870 };
     pagelength = l[papercode];
     if ( lowres ) pagelength /= 2;
   }
   if ( pagewidth == 0 ) {
     int w[] = { 0, 4190, 4800, 4800, 4800, 4138, 2202 };
     pagewidth = w[papercode];
     if ( lowres ) pagewidth /= 2;
     if ( highres ) pagewidth *= 2;
   }

   /* in mono mode, it's assumed we have a grey map specifiying
      the current page plane */
   if ( mono && !pageplane )
     pm_error("-mono is only allowed in page mode");

   /* if we're banding, subtract 37 (the one's for luck!)  from lfadj */
   if ( banding ) lfadj -= 37 ;

   if ( black ) { numrows = 1; }
   ppm_readppminit( ifd, &cols, &rows, &maxval,&format );

   /* if maxval is 1, then no point in dithering... */
   if ( maxval == 1 ) dithering = 0;

   if ( highres && !alreadyhigh ) cols *= 2 ;
   if (ppmout) ppm_writeppminit(stdout,cols+(xshift<0?xshift:0),
				rows+(yshift<0?yshift:0),maxval,0);
   if ( *greyfile ) {
     gfd = pm_openr(greyfile);
     if ( gfd == 0 ) pm_error("couldn't open greyfile");
     pgm_readpgminit(gfd,&gcols,&grows,(gray *)&gmaxval,&gformat);
     if ( highres && !alreadyhigh ) gcols *= 2;
     if ( gcols != cols || grows != rows || gmaxval != maxval )
       pm_error("greyfile doesn't match colour file");
     gpixels = pgm_allocrow(cols);
   }
   ithresh = FS_SCALE*fthresh;
   icdot = FS_SCALE*cdot;
   imdot = FS_SCALE*mdot;
   iydot = FS_SCALE*ydot;
   ikdot = FS_SCALE*kdot;
   pixels = ppm_allocrow(cols);

   /* limit checks */
   if (cols > (highres?2:1)*PCL_MAXWIDTH || rows > PCL_MAXHEIGHT)
      pm_error("image too large; reduce with ppmscale");
   if (maxval > PCL_MAXVAL)
      pm_error("colour range too large; reduce with ppmcscale");


   bpp = 8;
   bpp = (cols+bpp-1)/bpp;      /* bytes per row */

   if ((inrow = (char *)malloc((unsigned)bpp)) == NULL ||
       (outrow = (char *)malloc((unsigned)bpp*2)) == NULL ||
       (runcnt = (/*signed*/ char *)malloc((unsigned)bpp)) == NULL)
      pm_error("can't allocate space for row", 0,0,0,0,0);

   /* alloc bitrows for rgb -> cmyk */
   blackbits = (unsigned char *)malloc(cols);
   cyanbits = (unsigned char *)malloc(cols);
   magentabits = (unsigned char *)malloc(cols);
   yellowbits = (unsigned char *)malloc(cols);


   /* Initialize Floyd-Steinberg error vectors. */
   thiscerr = (long*) pm_allocrow( cols + 2, sizeof(long) );
   nextcerr = (long*) pm_allocrow( cols + 2, sizeof(long) );
   thismerr = (long*) pm_allocrow( cols + 2, sizeof(long) );
   nextmerr = (long*) pm_allocrow( cols + 2, sizeof(long) );
   thisyerr = (long*) pm_allocrow( cols + 2, sizeof(long) );
   nextyerr = (long*) pm_allocrow( cols + 2, sizeof(long) );
   thiskerr = (long*) pm_allocrow( cols + 2, sizeof(long) );
   nextkerr = (long*) pm_allocrow( cols + 2, sizeof(long) );
   srandom( (int) ( time( 0 ) ^ getpid( ) ) );
   /** inital errors should not be too large */
   for ( col = 0; col < cols + 2; ++col ) {
     thiscerr[col] = (random( ) % ( FS_SCALE * 2 ) - FS_SCALE)/8; /**/
     thismerr[col] = (random( ) % ( FS_SCALE * 2 ) - FS_SCALE)/8; /**/
     thisyerr[col] = (random( ) % ( FS_SCALE * 2 ) - FS_SCALE)/8; /**/
     thiskerr[col] = (random( ) % ( FS_SCALE * 2 ) - FS_SCALE)/8; /**/
     /* (random errors in [-1 .. 1]) */
   }
   fs_direction = 1;

   /* compute initgamma table x -> x^1.5 */
   if ( initgam == 0 ) initgam = ((halftone||fsdith)?(highres?1.8:1.6):1.2);
   for ( i = 0 ; i < 256 ; i++ ) {
     double ii = i/255.0;
     if ( initgam > 0.0 ) {
       ii = pow(ii,initgam);
     } else {
       ii = 1.0 - pow(1.0-ii,-initgam);
     }
     initgamma[i] = floor(255.0*ii+0.5);
   }

   /* if we're halftoning, compute the colour conversion lookup table:
      maps the top six bits each of cmy to cmyk. */
   if ( dithering && (!*greyfile && !mono && !nocolcor)) {
     colconv = (unsigned char *)malloc(4*64*64*64);
     for (index = 0, ic = 0 ; ic < 64 ; ic++ )
       for ( im = 0 ; im < 64 ; im++ )
	 for ( iy = 0 ; iy < 64 ; iy++ ) {
	   c = ic << 2; m = im << 2 ; y = iy << 2; k = 0;
	   jcb_cmyk_ht(&c,&m,&y,&k);
	   colconv[index++] = c;
	   colconv[index++] = m;
	   colconv[index++] = y;
	   colconv[index++] = k;
	 } 
     if ( solidblack ) {
       /* map black to full black, no cmy */
       /* index is currently pointing after k val of top cmy */
       colconv[--index] = '\377';
       colconv[--index] = '\0';
       colconv[--index] = '\0';
       colconv[--index] = '\0';
     }
   }
   if ( 1 ) {
     int four[] = { 0, 2, 3, 1 };
     kht.cellsize = 12;
     kht.dither = build_dith(6,(int *)dithmat6dotlow,2,four);
   } else {
     /* try building up from a 4x4 line screen */
     int fourdith[] = {
       8*24, 10*24, 12*24, 14*24,
       0*20,  2*20,  1*20,  3*20,
         4*20,  6*20,  5*20,  7*20,
       13*24, 11*24,  9*24, 15*24,
     } ;
     int four[] = { 0, 2, 3, 1 };
     kht.cellsize = 8;
     kht.dither = build_dith(4,fourdith,2,four);
     /* try different screen angles */
     kht.C.x = 63;
     kht.C.y = -16;
     kht.C.z = 65;
     kht.M.x = 63;
     kht.M.y = 16;
     kht.M.z = 65;
     kht.K.x = 4;
     kht.K.y = 3;
     kht.K.z = 5;
     kht.Y.x = 4;
     kht.Y.y = 3;
     kht.Y.z = 5;
   }
   /* write PCL header */
   
   /* in page plane mode, all settings carry through, and we
      we're still in data transfer mode (except for C...)
   */
   /* if ( !noreset && !(pageplane && !pageC)) printf("\033%%\200A"); */ /* select RGL mode */
   /* I hope this is always harmless! */
   if ( !ppmout ) printf("\033%%\200A"); /* select RGL mode */
   if ( !ppmout && (black || spotmode || ! pageplane || firstplane )) {/**/
     /* resolution defaults to 600 dpi */
     /* media type and fineness code */
     printf("\033*t%cR",highres?4:lowres?2:3);
     printf("\033&l%c%cM",mediacode,finecode); 
     printf("\033&l%c%cA",papercode,0); /* select paper size */
     /* page length, if set */
     if ( pagelength ) printf("\033&l%c%cP",pagelength%256,pagelength/256);
     /* page width, if given by user */
     if ( pagewidth ) printf("\033&l%c%cM",pagewidth%256,pagewidth/256);
     /* line feed adjustment */
     if ( lfadj != 0 ) { printf("\033\032%c%cL",(char) lfadj,0); }
     /* print head drive signal adjustment */
     if ( phadj != 0 ) { printf("\033\032%c%cV",(char) phadj,0); }

     if ( pageplane ) {
       printf("\033*r%cU",spotmode?7:4); /* selct YMCK/$col page plane printing */
     } else {
       printf("\033*r%cU",black?0:spotmode?5:2); /* select YMCK or K raster plane printing, or 4color page plane, or 4 color raster */
     }
     /* now the x and y offsets (if positive; if negative, we have
	to deal with them) */
     if (xshift > 0 )
       printf("\033&a%c%cL", xshift%256, xshift/256);
     if ( yshift > 0 ) 
       printf("\033&l%c%cE", yshift%256,yshift/256);
     /* if spot follow, disable curl correctoin */
     if ( !curlcorrection && (spotfollows || nocurl) ) printf("\033\032%c%cC",1,0);
     /* start raster graphics at CAP */
     printf("\033*r%cA",0);
   }
   /* in page mode (non spot), select the appropriate cartridge */
   if ( !ppmout && pageplane && ! spotmode) {
     printf("\033\032%c%cr",
	    pageK ? 0 : pageC ? 1 : pageM ? 2 : 3,
	    (spotfollows || nocurl) ? 0 : (curlcorrection || lastplane)) ; /* curl corrn on last plane only */
   }
   /* in spot mode (pageplane), select cartridge, with curl corrn */
   if ( !ppmout && spotmode && pageplane) printf("\033\032%c%cc",pageC?0:pageM?1:pageY?2:pageK?3:0,!nocurl);

   /* Now we process the rows */

   hrhalf = hrtoggle = 0 ;

   /* when we're skipping the rows that the printer's built-in
      banding code deals with, we have to know what more we're in,
      since for some bizarre reason the printer behaves differently
      for different modes. (Well, probably to avoid banding in the same place
      for each colour! 
      Recall the way banding works: 4 rows are printed by both the upper and lower
      print pass. So there are 232 rows printed in one pass only; the next 4 rows 
      are printed; the head advances 236 lines and the 4 rows are printed again;
      and we repeat. For the very first row, of course, the top four can only be
      printed by one pass. */

   bandingoffset = 0;
   if ( black ) {
     /* everything's easy: we offset by -4, cos the top of the first
	row is not subject to banding. */
     bandingoffset = -4 ;
   } else if ( pageplane && ! spotmode ) {
     /* in CMYK page mode, the offsets are: */
     if ( pageC ) bandingoffset = -4 ;
     if ( pageM ) bandingoffset = 176 ;
     if ( pageY ) bandingoffset = 56 ;
     if ( pageK ) bandingoffset = 56 ;
   }

   /* if we have a negative yshift, we need to skip rows */
   for ( row = 0 ; row < -yshift ; row++ ) {
     if ( highres && !alreadyhigh ) {
       ppm_readppmrow(ifd, pixels, cols/2, maxval,format );
     } else
       ppm_readppmrow(ifd, pixels, cols, maxval,format );
     if ( *greyfile ) {
       if ( highres && !alreadyhigh ) {
	  pgm_readpgmrow(gfd, gpixels, cols/2, maxval,format );
       } else
	 pgm_readpgmrow(gfd,gpixels,cols,maxval,gformat);
     }
   }
     
   /* row  is the number of the row in the image file.
      rowcounter is the number of the row relative to the start
      of the print pass.
   */
   for (row = (yshift >= 0 ? 0 : -yshift),
	  rowcounter = (yshift>=0 ? yshift : 0)+bandingoffset;
	row < rows;
	row++,rowcounter++) {

     if ( banding ) {
       /* skip rows for banding */
       while ( rowcounter >= 232 ) {
	 rowcounter -= 232;
	 printf("\033*b%c%cY",4,0);
       }
     }

     if ( highres && !alreadyhigh ) {
       ppm_readppmrow(ifd, pixels, cols/2, maxval,format );
       /* do the expansion */
       for ( i = cols/2 - 1 ; i >= 0 ; i-- ) 
	 pixels[2*i+1] = pixels[2*i] = pixels[i] ;
     } else
       ppm_readppmrow(ifd, pixels, cols, maxval,format );
     if ( *greyfile ) {
       if ( highres && !alreadyhigh ) {
	  pgm_readpgmrow(gfd, gpixels, cols/2, maxval,format );
	  /* do the expansion */
	  for ( i = cols/2 - 1 ; i >= 0 ; i-- ) 
	    gpixels[2*i+1] = gpixels[2*i] = gpixels[i] ;
       } else
	 pgm_readpgmrow(gfd,gpixels,cols,maxval,gformat);
     }

     /* Clear the bitvectors (we do this since in highres halftoning they
	keep state between passes, so we can't just set them each time. */
     memset(cyanbits,0,cols);
     memset(magentabits,0,cols);
     memset(yellowbits,0,cols);
     memset(blackbits,0,cols);

     if (fsdith)       for ( col = 0; col < cols + 2; ++col )
	nextcerr[col] = nextmerr[col] = nextyerr[col] = nextkerr[col]= 0;

     /* In high resolution halftoning, what do we do? My halftoning simply isn't
	up to halftoning on a non-square aspect ratio. So we pretend we have double
	vertical resolution as well; then when we output pixels, we make it on if
	both halves are on, and if only one half is on we alternately make it on or off.
	Crude. Will it work? We'll also use a goto to repeat the row. This code is getting
	more and more disgusting!
	*/

     /* this is the point we'll return to in highres halftoning */
   redo_halftone:
     /* if halftoning, initialize the halftone variables */
     if ( halftone ) {
       ht_init(&kht,K,row);
       ht_init(&kht,C,row);
       ht_init(&kht,M,row);
       ht_init(&kht,Y,row);
     } 
     /* calculate boolean kb,mb,cb,yb values for the row --
	then deal with each plane */
     /* for FS this is tedious, owing to boustrophedon */
     for (
	  (!fsdith || fs_direction) ? 
	    (col = 0, pixrow=pixels, gpixrow = gpixels, kbp = blackbits,mbp = magentabits,
	     cbp = cyanbits, ybp = yellowbits)
	    : (col = cols-1, pixrow=&pixels[col], gpixrow = &gpixels[col], kbp = &blackbits[col],mbp = &magentabits[col],
	     cbp = &cyanbits[col], ybp = &yellowbits[col]);
	   (!fsdith || fs_direction) ? (col < cols) : (col >= 0);
	   (!fsdith || fs_direction) ? (col++, pixrow++, gpixrow++, kbp++, cbp++, mbp++, ybp++)
	    : (col--, pixrow--, gpixrow--, kbp--, cbp--, mbp--, ybp--)
	  ) {
       r = PPM_GETR( *pixrow );
       g = PPM_GETG( *pixrow );
       b = PPM_GETB( *pixrow );
       k = 0; c = maxval - r ; m = maxval - g ; y = maxval - b ;
       if ( *greyfile ) k = maxval - *gpixrow;
       else if ( mono || black ) k = c ;
       if ( dithering ) {
	 /* if we are being fed CMYK data directly, that is, we're
	    in mono mode or we have a greyfile, it's inappropriate
	    to do any colour correction.
	    If we are in -black mode, we should just do gamma correction
	    */
	 if ( !*greyfile && !mono && !nocolcor ) {
	   if ( black ) {
	     if ( k < 255 ) k = initgamma[k]; /* ensure solid black */
	   } else {
	     /* if fullblack, just do black generation */
	     if ( fullblack ) {
	       k = c ; if ( m < k ) k = m ; if ( y < k ) k = y ;
	       m -= k ; c -= k ; y -= k ;
	     } else {
	       index = ((c & 0xFC) << 12) |
		 ((m & 0xFC) << 6) | (y & 0xFC);
	       c = colconv[index++];
	       m = colconv[index++];
	       y = colconv[index++];
	       k = colconv[index];
	     }
	   }
	 }
	 if ( fsdith ) {
	   if ( pageC ) {
	     /* Use Floyd-Steinberg errors to adjust actual color. */
	     sc = c*FS_BASIC_SCALE + thiscerr[col + 1];
	     /* new colours are either on or off */
	     
	     /* ar etc are booleans for on/off ; nr etc are the values
		used in error correction 
		*/
	     if ( sc < ithresh ) ac = 0 ; else ac = 1 ;
	     nc = icdot*ac;
	   }
	   if ( pageM ) {
	     sm = m*FS_BASIC_SCALE + thismerr[col + 1];
	     if ( sm < ithresh ) am = 0 ; else am = 1 ;
	     nm = imdot*am;
	   }
	   if (pageY) {
	     sy = y*FS_BASIC_SCALE + thisyerr[col + 1];
	     if ( sy < ithresh ) ay = 0 ; else ay = 1 ;
	     ny = iydot*ay;
	   }
	   if (pageK) {
	     sk = k*FS_BASIC_SCALE + thiskerr[col + 1];
	     if ( sk < ithresh ) ak = 0 ; else ak = 1 ;
	     nk = ikdot*ak;
	   }

	   /* Propagate Floyd-Steinberg error terms. */
	   if ( fs_direction )
	     {
	       if (pageC) {
		 err = ( sc - (long) nc );
		 thiscerr[col + 2] += ( err * 7 ) / 16;
		 nextcerr[col    ] += ( err * 3 ) / 16;
		 nextcerr[col + 1] += ( err * 5 ) / 16;
		 nextcerr[col + 2] += ( err     ) / 16;
	       }
	       if (pageM) {
		 err = ( sm - (long) nm );
		 thismerr[col + 2] += ( err * 7 ) / 16;
		 nextmerr[col    ] += ( err * 3 ) / 16;
		 nextmerr[col + 1] += ( err * 5 ) / 16;
		 nextmerr[col + 2] += ( err     ) / 16;
	       }
	       if (pageY) {
		 err = ( sy - (long) ny );
		 thisyerr[col + 2] += ( err * 7 ) / 16;
		 nextyerr[col    ] += ( err * 3 ) / 16;
		 nextyerr[col + 1] += ( err * 5 ) / 16;
		 nextyerr[col + 2] += ( err     ) / 16;
	       }
	       if (pageK) {
		 err = ( sk - (long) nk );
		 thiskerr[col + 2] += ( err * 7 ) / 16;
		 nextkerr[col    ] += ( err * 3 ) / 16;
		 nextkerr[col + 1] += ( err * 5 ) / 16;
		 nextkerr[col + 2] += ( err     ) / 16;
	       }
	     }
	   else
	     {
	       if (pageC) {
		 err = ( sc - (long) nc );
		 thiscerr[col    ] += ( err * 7 ) / 16;
		 nextcerr[col + 2] += ( err * 3 ) / 16;
		 nextcerr[col + 1] += ( err * 5 ) / 16;
		 nextcerr[col    ] += ( err     ) / 16;
	       }
	       if (pageM) {
		 err = ( sm - (long) nm );
		 thismerr[col    ] += ( err * 7 ) / 16;
		 nextmerr[col + 2] += ( err * 3 ) / 16;
		 nextmerr[col + 1] += ( err * 5 ) / 16;
		 nextmerr[col    ] += ( err     ) / 16;
	       }
	       if (pageY) {
		 err = ( sy - (long) ny );
		 thisyerr[col    ] += ( err * 7 ) / 16;
		 nextyerr[col + 2] += ( err * 3 ) / 16;
		 nextyerr[col + 1] += ( err * 5 ) / 16;
		 nextyerr[col    ] += ( err     ) / 16;
	       }
	       if (pageK) {
		 err = ( sk - (long) nk );
		 thiskerr[col    ] += ( err * 7 ) / 16;
		 nextkerr[col + 2] += ( err * 3 ) / 16;
		 nextkerr[col + 1] += ( err * 5 ) / 16;
		 nextkerr[col    ] += ( err     ) / 16;
	       }
	     }
	   *kbp = ak;
	   *cbp = ac;
	   *mbp = am;
	   *ybp = ay;
	 } else {
           if ( halftone  ) {
	      int t;
	   /* Standard halftoning:
	      The bit vectors have been cleared; we *add* the relevant value.
	      This is because in highres halftoning, our pixels have values
	      0,1,2.
	      */
	    
	      if (coldiag) {
		/* coldiag, leave unchanged */
		*kbp = k; *cbp = c; *mbp = m; *ybp = y;
	      } else {
		/* non standard screen angles to avoid banding on black */
		/* in low res mode, if the pixel value equals
		   the ht dot value+1, set the dot half the time.
		*/
#define LOWHALF (1 || highres || (hrtoggle ^= 1))
		if ( pageY ) {
		  /* rotate(row,col,&yhrow,&yhcol,4,3,5); */
		  t = y - ht_elt(&kht,Y);
		  *ybp +=  (t > 1) ? 1 : (t > 0) ? LOWHALF : 0 ;
		  ht_inc(&kht,Y);
		}
		if (pageK) {
		  /* rotate(row,col,&khrow,&khcol,4,3,5); */
		  t = k - ht_elt(&kht,K);
		  *kbp += (t > 1) ? 1 : (t > 0) ? LOWHALF : 0 ;
		  ht_inc(&kht,K);
		}
		if (pageC) {
		  /* rotate(row,-col,&chrow,&chcol,HT_C_X,HT_C_Y,HT_C_Z); */
		  t = c - ht_telt(&kht,C);
		  *cbp += (t > 1) ? 1 : (t > 0) ? LOWHALF : 0 ;
		  ht_inc(&kht,C);
		}
		if (pageM) {
		  /* rotate(row,-col,&mhrow,&mhcol,60,11,61); */
		  t = m - ht_telt(&kht,M);
		  *mbp += (t > 1) ? 1 : (t > 0) ? LOWHALF : 0 ;
		  ht_inc(&kht,M);
		}
	      }
	   } else {
	     /* dithering in the square pattern with unrotated screens */
	     if ( highres ) {
	       static int offset ;
	       /* set each halftone cell in a random orientation */
	       if ( row%4 == 0 && col%8 == 0 ) offset = 4*((rand()&0xFF)>>7) ;
	       if (pageK) *kbp = (k > hcluster[row%4][(col+4-offset)%8]);
	       if (pageC) *cbp = (c > hcluster[row%4][(col+offset)%8]);
	       if (pageM) *mbp = (m > hcluster[row%4][(col+offset)%8]);
	       if (pageY) *ybp = (y > hcluster[row%4][(col+offset)%8]);
	     } else {
	       if (pageK) *kbp = (k > cluster8[row%8][(col+4)%8]);
	       if (pageC) *cbp = (c > cluster8[row%8][col%8]);
	       if (pageM) *mbp = (m > cluster8[row%8][col%8]);
	       if (pageY) *ybp = (y > cluster8[row%8][col%8]);
	     }
	   }
	 }
       } else {
	 /* we assume full bgucr in this snap to primary model,
	    unless a separate greyfile is specified */
	 if ( (*greyfile && *gpixrow == 0)
	      || ( ! *greyfile 
		   && r == 0 && g == 0 && b == 0 ))
	   *kbp = 1 ; else *kbp = 0 ;
	 /* cyan is one if red is 0, unless black is 1 and no greyfile */
	 /* also, if mono, we assume r = g = b, so only look at red
	    for all the hues */
	 if ( r == 0
	      && (mono || *greyfile || *kbp == 0 ) )
	   *cbp = 1 ; else *cbp = 0 ;
	 if ( g == 0
	      && (mono || *greyfile || *kbp == 0 ) )
	   *mbp = 1 ; else *mbp = 0 ;
	 if ( b == 0
	      && (mono || *greyfile || *kbp == 0 ) )
	   *ybp = 1 ; else *ybp = 0 ;
       }
     } /* end of the column for loop */
     if (fsdith) {
       temperr = thiscerr;
       thiscerr = nextcerr;
       nextcerr = temperr;
       temperr = thismerr;
       thismerr = nextmerr;
       nextmerr = temperr;
       temperr = thisyerr;
       thisyerr = nextyerr;
       nextyerr = temperr;
       temperr = thiskerr;
       thiskerr = nextkerr;
       nextkerr = temperr;
       fs_direction = ! fs_direction;
     }
     /* if we're in highresolution halftoning mode we need either to redo
	the row, or fix up the values */
     /* #define HALFPIX ((rand() & 0xFF)>>7) */
#define HALFPIX (hrtoggle ^= 1) 

     if ( highres && halftone ) {
       if ( hrhalf ) {
	 /* we've just done the second half row; fix the values, and we're done */
	 if ( pageC ) {
	   for ( col = 0, cbp = cyanbits; col < cols ; col++, cbp++ )
	     if ( *cbp == 2 ) *cbp = 1 ;
	     else if ( *cbp ) *cbp = HALFPIX ;
	 }
	 if ( pageM ) {
	   for ( col = 0, mbp = magentabits; col < cols ; col++, mbp++ )
	     if ( *mbp == 2 ) *mbp = 1 ;
	     else if ( *mbp ) *mbp = HALFPIX ;
	 }
	 if ( pageY ) {
	   for ( col = 0, ybp = yellowbits; col < cols ; col++, ybp++ )
	     if ( *ybp == 2 ) *ybp = 1 ;
	     else if ( *ybp ) *ybp = HALFPIX ;
	 }
	 if ( pageK ) {
	   for ( col = 0, kbp = blackbits; col < cols ; col++, kbp++ )
	     if ( *kbp == 2 ) *kbp = 1 ;
	     else if ( *kbp ) *kbp = HALFPIX ;
	 }
	 hrhalf = 0 ;
       } else {
	 /* we've done the first half row, so we need to go round again */
	 hrhalf = 1 ;
	 goto redo_halftone;
       }
     }
     /* if ppmout, reassign the pixels */
     if (ppmout) {
       for ( col = 0 , pixrow = pixels ,
	       cbp = cyanbits, mbp = magentabits, ybp = yellowbits, kbp = blackbits;
	     col < cols ; col++, pixrow++, cbp++ , mbp++, ybp++, kbp++ )
	 if ( coldiag ) {
	   PPM_ASSIGN(*pixrow,
		      maxval-((pageC ? *cbp : 0) + (pageK ? *kbp: 0)),
		      maxval-((pageM ? *mbp : 0) + (pageK ? *kbp: 0)),
		      maxval-((pageY ? *ybp : 0) + (pageK ? *kbp: 0)));
	 } else if ( ppmdiag ) {
	   PPM_ASSIGN(*pixrow,
		      maxval - 70*(*cbp) - 140*(*kbp),
		      maxval - 70*(*mbp) - 140*(*kbp),
		      maxval - 70*(*ybp) - 140*(*kbp));
	 } else {
	   PPM_ASSIGN(*pixrow,
		      maxval*(!(pageK && *kbp) && !(pageC && *cbp)),
		      maxval*(!(pageK &&*kbp) && !(pageM && *mbp)),
		      maxval*(!(pageK &&*kbp) && !(pageY && *ybp))) ; 
	 }
     }
     
     /* are we printing this row (if only odd or even)? */
     printrow = 1;
     if ( evenrowsonly && (row & 1) ) printrow = 0;
     if ( oddrowsonly && ! (row & 1) )  printrow = 0;
     /* if we have a negative xshift, we need to skip columns */
     skipcols = 0;
     if ( xshift < 0 ) skipcols = -xshift ;
     /* send K C M Y */
     if ( pageK && printrow) {
       putbits(blackbits+skipcols,cols-skipcols);
     } else if ( !black && ! pageplane ) {
       /* we need to skip the row */
       putbits(blackbits,0);
     }
     if ( pageC && printrow ) {
       putbits(cyanbits+skipcols,cols-skipcols);
     } else if ( !black && ! pageplane ) {
       /* we need to skip the row */
       putbits(cyanbits,0);
     }

     if ( pageM  && printrow) {
       putbits(magentabits+skipcols,cols-skipcols);
     } else if ( !black && ! pageplane ) {
       /* we need to skip the row */
       putbits(magentabits,0);
     }

     if ( pageY  && printrow) {
       putbits(yellowbits+skipcols,cols-skipcols);
     } else if ( !black && ! pageplane ) {
       /* we need to skip the row */
       putbits(yellowbits,0);
     }

     if (ppmout) ppm_writeppmrow(stdout,pixels+(sizeof(pixel)*skipcols),
				 cols-skipcols,maxval,0);
   }
   if ( ! ppmout ) {
     /* In page mode (non spot), we stay in data transfer mode until the K page, and
	we go to top of form */
     if ( pageplane && ! spotmode && ! lastplane ) {
       if ( !noreset) printf("\033\032%c%c\014",0,0);
     } else {
       if ( spotfollows ) printf("\033\032%c%c\014",0,0); /* tof */
       if ( !ffonly ) printf("\033*rC"); /* end raster graphics */
       if ( !noreset && !spotfollows ) {
	 if ( ffonly ) {
	   printf("\014"); /* formfeed */
	 } else {
	   printf("\014\033%%%cX\033e",0); /* full printer reset */
	 }
       }
     }
   }
   exit(0);
}
